@@ -12,9 +12,11 @@ import deepr.jobs
 import deepr.layers
 import deepr.metrics
 import deepr.optimizers
+import deepr.predictors
 import deepr.prepros
 import deepr.readers
 import deepr.utils
+import deepr.vocab
 from deepr.config import parse_config, from_config
 from deepr.utils import Field
 from deepr.version import __author__, __version__
