@@ -1,4 +1,4 @@
 # pylint: disable=unused-import,missing-docstring
 
 from deepr.example.jobs.build_dataset import BuildDataset
-from deepr.example.jobs.predict import Predict
+from deepr.example.jobs.predict import PredictProto, PredictSavedModel
